#! /usr/bin/python
# Copyright Notice:
# Copyright 2019-2025 DMTF. All rights reserved.
# License: BSD 3-Clause License. For full text see link: https://github.com/DMTF/Redfish-Tacklebox/blob/main/LICENSE.md

"""
Redfish Power Reset

File : rf_power_reset.py

Brief : This script uses the redfish_utilities module to perform a reset of the system
"""

import argparse
import datetime
import logging
import redfish
import redfish_utilities
import traceback
import sys
from redfish.messages import RedfishPasswordChangeRequiredError

# Get the input arguments
argget = argparse.ArgumentParser(description="A tool to perform a power/reset operation of a system")
argget.add_argument("--user", "-u", type=str, required=True, help="The user name for authentication")
argget.add_argument("--password", "-p", type=str, required=True, help="The password for authentication")
argget.add_argument("--rhost", "-r", type=str, required=True, help="The address of the Redfish service (with scheme)")
argget.add_argument("--system", "-s", type=str, help="The ID of the system to reset")
argget.add_argument(
    "--type", "-t", type=str, help="The type of power/reset operation to perform", choices=redfish_utilities.reset_types
)
argget.add_argument(
    "--info", "-info", action="store_true", help="Indicates if reset and power information should be reported"
)
argget.add_argument("--debug", action="store_true", help="Creates debug file showing HTTP traces and exceptions")
args = argget.parse_args()

if args.debug:
    log_file = "rf_power_reset-{}.log".format(datetime.datetime.now().strftime("%Y-%m-%d-%H%M%S"))
    log_format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    logger = redfish.redfish_logger(log_file, log_format, logging.DEBUG)
    logger.info("rf_power_reset Trace")

# Set up the Redfish object
redfish_obj = None
try:
    redfish_obj = redfish.redfish_client(
        base_url=args.rhost, username=args.user, password=args.password, timeout=15, max_retry=3
    )
    redfish_obj.login(auth="session")
except RedfishPasswordChangeRequiredError:
    redfish_utilities.print_password_change_required_and_logout(redfish_obj, args)
    sys.exit(1)
except Exception:
    raise

exit_code = 0
try:
    if args.info:
        system_info = redfish_utilities.get_system(redfish_obj, args.system)
        reset_uri, reset_parameters = redfish_utilities.get_system_reset_info(redfish_obj, args.system, system_info)
        printed_reset_types = False
        for param in reset_parameters:
            if param["Name"] == "ResetType" and "AllowableValues" in param:
                print("Supported reset types: {}".format(", ".join(param["AllowableValues"])))
                printed_reset_types = True
        if not printed_reset_types:
            print("No reset information found")
        if "PowerState" in system_info.dict:
            print("Current power state: {}".format(system_info.dict["PowerState"]))
        else:
            print("No power state information found")
    else:
        print("Resetting the system...")
        response = redfish_utilities.system_reset(redfish_obj, args.system, args.type)
        response = redfish_utilities.poll_task_monitor(redfish_obj, response)
        redfish_utilities.verify_response(response)
except Exception as e:
    if args.debug:
        logger.error("Caught exception:\n\n{}\n".format(traceback.format_exc()))
    exit_code = 1
    print(e)
finally:
    # Log out
    redfish_utilities.logout(redfish_obj)
sys.exit(exit_code)



#! /usr/bin/python
# Copyright Notice:
# Copyright 2019-2025 DMTF. All rights reserved.
# License: BSD 3-Clause License. For full text see link: https://github.com/DMTF/Redfish-Tacklebox/blob/main/LICENSE.md

"""
Messages Module

File : messages.py

Brief : This file contains the definitions and functionalities for interacting
        with Messages for a given Redfish service
"""

from redfish.messages import (
    get_messages_detail,
    get_error_messages,
    search_message,
    RedfishPasswordChangeRequiredError,
    RedfishOperationFailedError,
)


def verify_response(response):
    """
    Verifies a response and raises an exception if there was a failure

    Args:
        response: The response to verify
    """

    if response.status >= 400:
        messages_detail = get_messages_detail(response)
        exception_string = get_error_messages(messages_detail)
        message_item = search_message(messages_detail, "Base", "PasswordChangeRequired")
        if message_item is not None:
            raise RedfishPasswordChangeRequiredError(
                "Operation failed: HTTP {}\n{}".format(response.status, exception_string),
                message_item["MessageArgs"][0],
            )
        else:
            raise RedfishOperationFailedError("Operation failed: HTTP {}\n{}".format(response.status, exception_string))

    return


def print_error_payload(response):
    """
    Prints an error payload, which can also be used for action responses

    Args:
        response: The response to print
    """

    try:
        print(get_error_messages(response))
    except Exception:
        # No response body
        if response.status >= 400:
            print("Failed")
        else:
            print("Success")



#! /usr/bin/python
# Copyright Notice:
# Copyright 2019-2025 DMTF. All rights reserved.
# License: BSD 3-Clause License. For full text see link: https://github.com/DMTF/Redfish-Tacklebox/blob/main/LICENSE.md

"""
Update Module

File : update.py

Brief : This file contains the definitions and functionalities for interacting
        with the UpdateService for a given Redfish service
"""

import json
import os
import errno
from .collections import get_collection_members
from .messages import verify_response
from enum import Enum


class RedfishUpdateServiceNotFoundError(Exception):
    """
    Raised when the update service or an update action cannot be found
    """

    pass


class RedfishFirmwareInventoryNotFoundError(Exception):
    """
    Raised when the update service does not contain firmware inventory
    """

    pass


class operation_apply_times(Enum):
    """
    Values for operation apply time settings
    """

    IMMEDIATE = "Immediate"
    ON_RESET = "OnReset"
    AT_MAINTENANCE_WINDOW_START = "AtMaintenanceWindowStart"
    AT_MAINTENANCE_WINDOW_ON_RESET = "InMaintenanceWindowOnReset"
    ON_START_UPDATE_REQUEST = "OnStartUpdateRequest"

    def __str__(self):
        return self.value


def get_simple_update_info(context):
    """
    Locates the SimpleUpdate action and collects its information

    Args:
        context: The Redfish client object with an open session

    Returns:
        The URI of the SimpleUpdate action
        A list of parameter requirements from the action info
    """

    # Get the update service
    update_service = get_update_service(context)

    # Check that there is a SimpleUpdate action
    if "Actions" not in update_service.dict:
        raise RedfishUpdateServiceNotFoundError("Service does not support SimpleUpdate")
    if "#UpdateService.SimpleUpdate" not in update_service.dict["Actions"]:
        raise RedfishUpdateServiceNotFoundError("Service does not support SimpleUpdate")

    # Extract the info about the SimpleUpdate action
    simple_update_action = update_service.dict["Actions"]["#UpdateService.SimpleUpdate"]
    simple_update_uri = simple_update_action["target"]

    if "@Redfish.ActionInfo" not in simple_update_action:
        # No action info; need to build this manually based on other annotations

        # Default parameter requirements
        simple_update_parameters = [
            {"Name": "ImageURI", "Required": True, "DataType": "String"},
            {
                "Name": "TransferProtocol",
                "Required": False,
                "DataType": "String",
                "AllowableValues": ["CIFS", "FTP", "SFTP", "HTTP", "HTTPS", "NSF", "SCP", "TFTP", "OEM", "NFS"],
            },
            {"Name": "Targets", "Required": False, "DateType": "StringArray"},
            {"Name": "Username", "Required": False, "DataType": "String"},
            {"Name": "Password", "Required": False, "DataType": "String"},
        ]

        # Get the allowable values from annotations
        for param in simple_update_parameters:
            if param["Name"] + "@Redfish.AllowableValues" in simple_update_action:
                param["AllowableValues"] = simple_update_action[param["Name"] + "@Redfish.AllowableValues"]
    else:
        # Get the action info and its parameter listing
        action_info = context.get(simple_update_action["@Redfish.ActionInfo"])
        simple_update_parameters = action_info.dict["Parameters"]

    return simple_update_uri, simple_update_parameters


def simple_update(context, image_uri, protocol=None, targets=None, username=None, password=None, apply_time=None):
    """
    Performs a SimpleUpdate request

    Args:
        context: The Redfish client object with an open session
        image_uri: The image URI for the update
        protocol: The transfer protocol for the update
        targets: The targets receiving the update
        username: The username for retrieving the update for the given URI
        password: The password for retrieving the update for the given URI
        apply_time: The apply time for the update

    Returns:
        The response from the request
    """

    # Get the SimpleUpdate info
    uri, params = get_simple_update_info(context)

    # Build the request body
    body = {"ImageURI": image_uri}
    if protocol is not None:
        body["TransferProtocol"] = protocol
    if targets is not None:
        body["Targets"] = targets
    if username is not None:
        body["Username"] = username
    if password is not None:
        body["Password"] = password
    if apply_time is not None:
        body["@Redfish.OperationApplyTime"] = apply_time.value

    response = context.post(uri, body=body)
    verify_response(response)
    return response


def get_size(file_path, unit="bytes"):
    """
    Determines the size of a local file

    Args:
        file_path: The path to the file
        unit: The units to apply to the return value

    Returns:
        The size of the file in the specified units
    """

    file_size = os.path.getsize(file_path)
    exponents_map = {"bytes": 0, "kb": 1, "mb": 2, "gb": 3}
    if unit not in exponents_map:
        raise ValueError("Must select from ['bytes', 'kb', 'mb', 'gb']")
    else:
        size = file_size / 1024 ** exponents_map[unit]
        return round(size, 3)


def multipart_push_update(context, image_path, targets=None, timeout=None, apply_time=None):
    """
    Performs an HTTP Multipart push update request

    Args:
        context: The Redfish client object with an open session
        image_path: The filepath to the image for the update
        targets: The targets receiving the update
        timeout: The timeout to apply to the update
        apply_time: The apply time for the update

    Returns:
        The response from the request
    """

    # Ensure the file exists
    if os.path.isfile(image_path) is False:
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), image_path)

    # If no update is specified, determine an appropriate timeout to apply
    if timeout is None:
        """TODO: See what a 'reasonable' timeout is when accounting for slow networks
        for now, keeping the timeout conservative (2 seconds per MB)
        timeout = 5
        file_size = get_size( image_path, "mb" )

        if file_size >= 16:
            timeout = math.ceil( ( 5 / 16 ) * file_size )
        """
        timeout = 30
        file_size = get_size(image_path, "mb")
        if file_size >= 15:
            timeout = 2 * file_size

    # Get the update service
    update_service = get_update_service(context)
    if "MultipartHttpPushUri" not in update_service.dict:
        raise RedfishUpdateServiceNotFoundError("Service does not support MultipartHttpPushUri")

    # Build the request body
    update_parameters = {}
    if targets is not None:
        update_parameters["Targets"] = targets
    if apply_time is not None:
        update_parameters["@Redfish.OperationApplyTime"] = apply_time.value
    body = {
        "UpdateParameters": (None, json.dumps(update_parameters), "application/json"),
        "UpdateFile": (image_path.split(os.path.sep)[-1], open(image_path, "rb"), "application/octet-stream"),
    }

    response = context.post(
        update_service.dict["MultipartHttpPushUri"],
        body=body,
        headers={"Content-Type": "multipart/form-data"},
        timeout=timeout,
        max_retry=3,
    )
    verify_response(response)
    return response


def get_firmware_inventory(context):
    """
    Finds the firmware inventory and returns its contents

    Args:
        context: The Redfish client object with an open session

    Returns:
        An array of dictionaries of the firmware inventory members
    """

    # Get the update service
    update_service = get_update_service(context)

    # Check that there is a firmware inventory collection
    if "FirmwareInventory" not in update_service.dict:
        raise RedfishFirmwareInventoryNotFoundError("Service does not have a firmware inventory")

    return get_collection_members(context, update_service.dict["FirmwareInventory"]["@odata.id"])


def print_software_inventory(software_list, details=False, use_id=False):
    """
    Prints the software inventory list into a table

    Args:
        software_list: The certificate list to print
        details: True to print all the detailed info
        use_id: Indicates whether to print names from 'Id' property values
    """

    software_line_format = "  {:40s} | {}"
    software_details_line_format = "  {:40s} | {}: {}"
    details_properties = ["Manufacturer", "SoftwareId", "ReleaseDate"]

    # Go through each software element
    for software in software_list:
        if use_id:
            name = software["Id"]
        else:
            name = software["Name"]
        version = software.get("Version", "No version found")
        if details:
            print(software_details_line_format.format(name, "Version", version))
            for detail in details_properties:
                if detail in software:
                    print(software_details_line_format.format("", detail, software[detail]))
        else:
            print(software_line_format.format(name, version))
    print("")


def get_update_service(context):
    """
    Locates and gets the UpdateService resource

    Args:
        context: The Redfish client object with an open session

    Returns:
        The UpdateService resource
    """

    # Get the Service Root to find the Update Service
    service_root = context.get("/redfish/v1/")
    if "UpdateService" not in service_root.dict:
        # No Update Service
        raise RedfishUpdateServiceNotFoundError("Service does not have an UpdateService")

    return context.get(service_root.dict["UpdateService"]["@odata.id"])



#! /usr/bin/python
# Copyright Notice:
# Copyright 2019-2025 DMTF. All rights reserved.
# License: BSD 3-Clause License. For full text see link: https://github.com/DMTF/Redfish-Tacklebox/blob/main/LICENSE.md

"""
Redfish Discover

File : rf_discover.py

Brief : This script uses the redfish module to discover Redfish services
"""

import argparse
import re
import redfish
import sys

# No arguments, but having help text is useful
argget = argparse.ArgumentParser(description="A tool to discover Redfish services")
args = argget.parse_args()

# Invoke the discovery routine for SSDP and print the responses
services = redfish.discover_ssdp()
if len(services) == 0:
    print("No Redfish services discovered")
    sys.exit(1)
else:
    print("Redfish services:")

# Go through each discovered service and print out basic info
for service in services:
    # Try to get the Product property from the service root for each service
    # If found, print the UUID, service root pointer, and the Product
    # If not, just print the UUID and service root pointer
    try:
        # Need to strip off /redfish/v1 from the SSDP response to use the URL with the library
        groups = re.search(r"^(.+)\/redfish\/v1\/?$", services[service])
        url = groups.group(1)
        redfish_obj = redfish.redfish_client(base_url=url, timeout=15, max_retry=3)
        print("{}: {} ({})".format(service, services[service], redfish_obj.root["Product"]))
    except Exception:
        print("{}: {}".format(service, services[service]))

sys.exit(0)



#! /usr/bin/python
# Copyright Notice:
# Copyright 2019-2025 DMTF. All rights reserved.
# License: BSD 3-Clause License. For full text see link: https://github.com/DMTF/Redfish-Tacklebox/blob/main/LICENSE.md

from setuptools import setup
from setuptools import Command as _Command
from codecs import open
import os


class Pyinstaller(_Command):
    description: "Pyinstaller"
    user_options = []

    def __init__(self, dist, **kw):
        super().__init__(dist, **kw)
        self.flags = ""
        self.scripts = []
        self.packages = []
        if not os.path.exists("./dist"):
            os.mkdir("dist")
        if not os.path.exists("./spec"):
            os.mkdir("spec")

    def initialize_options(self):
        self.flags = ""
        self.scripts = []
        self.packages = []

    def finalize_options(self):
        self.flags = "--specpath ./spec"
        self.scripts = self.distribution.scripts
        self.packages = self.distribution.packages
        for package in self.packages:
            self.flags = "{} --collect-all {}".format(self.flags, package)

    def pyinstaller(self, target):
        if os.system("pyinstaller --onefile {} {}".format(target, self.flags)):
            raise Exception("PyInstaller failed!")

    def run(self):
        for scripts in self.scripts:
            self.pyinstaller(scripts)


with open("README.md", "r", "utf-8") as f:
    long_description = f.read()

setup(
    name="redfish_utilities",
    version="3.3.9",
    description="Redfish Utilities",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="DMTF, https://www.dmtf.org/standards/feedback",
    license='BSD 3-clause "New" or "Revised License"',
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
        "Topic :: Communications",
    ],
    keywords="Redfish",
    url="https://github.com/DMTF/Redfish-Tacklebox",
    packages=["redfish_utilities"],
    scripts=[
        "scripts/rf_accounts.py",
        "scripts/rf_assembly.py",
        "scripts/rf_bios_settings.py",
        "scripts/rf_boot_override.py",
        "scripts/rf_certificates.py",
        "scripts/rf_diagnostic_data.py",
        "scripts/rf_discover.py",
        "scripts/rf_event_service.py",
        "scripts/rf_firmware_inventory.py",
        "scripts/rf_licenses.py",
        "scripts/rf_logs.py",
        "scripts/rf_manager_config.py",
        "scripts/rf_power_equipment.py",
        "scripts/rf_power_reset.py",
        "scripts/rf_raw_request.py",
        "scripts/rf_sel.py",
        "scripts/rf_sensor_list.py",
        "scripts/rf_sys_inventory.py",
        "scripts/rf_test_event_listener.py",
        "scripts/rf_thermal_equipment.py",
        "scripts/rf_update.py",
        "scripts/rf_virtual_media.py",
    ],
    install_requires=["redfish>=3.2.1", "XlsxWriter>=1.2.7", "requests"],
    cmdclass={"pyinstaller": Pyinstaller},
)



#! /usr/bin/python
# Copyright Notice:
# Copyright 2019-2025 DMTF. All rights reserved.
# License: BSD 3-Clause License. For full text see link: https://github.com/DMTF/Redfish-Tacklebox/blob/main/LICENSE.md

"""
Licenses Module

File : licenses.py

Brief : This file contains the definitions and functionalities for managing
        licenses on a Redfish service
"""

import base64
import os
from .collections import get_collection_ids
from .messages import verify_response


class RedfishLicenseServiceNotFoundError(Exception):
    """
    Raised when the license service cannot be found
    """

    pass


class RedfishLicenseCollectionNotFoundError(Exception):
    """
    Raised when the license collection cannot be found
    """

    pass


class RedfishLicenseNotFoundError(Exception):
    """
    Raised when a specific license cannot be found
    """

    pass


class RedfishInstallLicenseNotFoundError(Exception):
    """
    Raised when the license service does not contain the Install action
    """

    pass


def get_licenses(context):
    """
    Collects license information from a Redfish service

    Args:
        context: The Redfish client object with an open session

    Returns:
        A list containing all licenses
    """

    license_list = []
    license_collection = get_license_collection(context)

    # Get the identifiers of the collection
    license_col = get_collection_ids(context, license_collection)

    # Get each member and add it to the response list
    for license_id in license_col:
        license = context.get(license_collection + "/" + license_id)
        verify_response(license)
        license_list.append(license.dict)

    return license_list


def print_licenses(license_list, details=False):
    """
    Prints the license list into a table

    Args:
        license_list: The license list to print
        details: True to print all the detailed info
    """

    license_format = "  {:30s} | {}"
    license_format_detail = "  {:30s} | {}: {}"
    info_properties = [
        {"Format": "{}", "Property": "EntitlementId"},
        {"Format": "Installed on {}", "Property": "InstallDate"},
        {"Format": "Expires on {}", "Property": "ExpirationDate"},
    ]
    detail_list = [
        "Description",
        "LicenseType",
        "LicenseOrigin",
        "Removable",
        "Manufacturer",
        "SKU",
        "PartNumber",
        "SerialNumber",
        "AuthorizationScope",
        "MaxAuthorizedDevices",
        "RemainingUseCount",
    ]

    print("")
    print(license_format.format("License", "Details"))

    # Go through each license
    for license in license_list:
        info_list = []
        # Build the general info string
        for info in info_properties:
            if info["Property"] in license:
                info_string = info["Format"].format(license[info["Property"]])
            else:
                if info["Property"] == "EntitlementId":
                    # Fallback to ensure there is always something to show
                    info_string = "License " + license["Id"]
            info_list.append(info_string)

        # Print the license info
        print(license_format.format(license["Id"], ", ".join(info_list)))

        # Print details if requested
        if details:
            for detail in detail_list:
                if detail in license:
                    print(license_format_detail.format("", detail, license[detail]))

    print("")


def install_license(context, license_path):
    """
    Installs a new license

    Args:
        context: The Redfish client object with an open session
        license_path: The filepath or URI of the license to install

    Returns:
        The response of the operation
    """

    # Get the license service
    license_service = get_license_service(context)

    # Determine which installation method to use based on the provided license path
    if os.path.isfile(license_path):
        # Local file; perform via a POST to the license collection
        if "Licenses" not in license_service:
            raise RedfishLicenseCollectionNotFoundError("The license service does not contain a license collection")
        install_uri = license_service["Licenses"]["@odata.id"]

        # Read in the file and convert it to a Base64-encoded string
        with open(license_path, "rb") as file:
            license_file = file.read()
        payload = {"LicenseString": base64.b64encode(license_file).decode("utf-8")}
    else:
        # Remote file; perform via a POST to the Install action
        if "Actions" not in license_service:
            raise RedfishInstallLicenseNotFoundError("The license service does not contain actions")
        if "#LicenseService.Install" not in license_service["Actions"]:
            raise RedfishInstallLicenseNotFoundError("The license service does not contain the Install action")
        install_uri = license_service["Actions"]["#LicenseService.Install"]["target"]

        payload = {"LicenseFileURI": license_path}

    # Install the license
    response = context.post(install_uri, body=payload)
    verify_response(response)
    return response


def delete_license(context, license_id):
    """
    Deletes a license

    Args:
        context: The Redfish client object with an open session
        license_id: The identifier of the license to delete

    Returns:
        The response of the operation
    """

    # Get the identifiers of the collection
    license_collection = get_license_collection(context)
    avail_licenses = get_collection_ids(context, license_collection)
    if license_id not in avail_licenses:
        raise RedfishLicenseNotFoundError(
            "License service does not contain the license '{}'; available licenses: {}".format(
                license_id, ", ".join(avail_licenses)
            )
        )

    # Delete the requested license
    response = context.delete(license_collection + "/" + license_id)
    verify_response(response)
    return response


def get_license_service(context):
    """
    Collects the license service information from a Redfish service

    Args:
        context: The Redfish client object with an open session

    Returns:
        An object containing information about the license service
    """

    # Get the service root to find the license service
    service_root = context.get("/redfish/v1/")
    if "LicenseService" not in service_root.dict:
        # No event service
        raise RedfishLicenseServiceNotFoundError("Service does not contain a license service")

    # Get the license service
    license_service = context.get(service_root.dict["LicenseService"]["@odata.id"])
    verify_response(license_service)
    return license_service.dict


def get_license_collection(context):
    """
    Finds the license collection for the Redfish service

    Args:
        context: The Redfish client object with an open session

    Returns:
        The URI for the license collection
    """

    # Get the license service to find the license collection
    license_service = get_license_service(context)
    if "Licenses" not in license_service:
        # No license collection
        raise RedfishLicenseCollectionNotFoundError("Service does not contain a license collection")

    return license_service["Licenses"]["@odata.id"]
